from typing import Any
from typing import Dict


class OptionEventEquityComponent(object):

    def __init__(
        self,
        component_id: str,
        symbol: str,
        side: str,
        instrument_url: str,
        price: float,
        quantity: float,
    ) -> None:
        self._component_id = component_id
        self._symbol = symbol
        self._side = side
        self._instrument_url = instrument_url
        self._price = price
        self._quantity = quantity


def option_event_equity_component_from_data(data: Dict[str, Any]) -> OptionEventEquityComponent:
    return OptionEventEquityComponent(
        component_id=data['id'],
        symbol=data['symbol'],
        side=data['side'],
        instrument_url=data['instrument'],
        price=float(data['price']),
        quantity=float(data['quantity']),
    )



import json
from os import environ
from time import sleep
from time import time
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Set
from urllib.parse import urlencode
from urllib.request import urlopen

from alphavantage.pricehistory.pricehistory import price_history_from_data
from alphavantage.pricehistory.pricehistory import PriceHistory


class AlphaVantageClient(object):

    def __init__(self) -> None:
        self._cool_down_window = self._get_cool_down_window()
        self._last_request_time: Optional[float] = None

    def get_price_history(
        self,
        symbol: str,
        time_window: str,
        output_size: str,
    ) -> PriceHistory:
        symbol = symbol.upper()

        time_window = self._check_str_parameter_in_range(
            parameter_name='time_window',
            parameter_value=time_window,
            allowed_values={'daily', 'weekly', 'monthly'},
            enforce_upper_case=True,
        )

        output_size = self._check_str_parameter_in_range(
            parameter_name='output_size',
            parameter_value=output_size,
            allowed_values={'compact', 'full'},
            enforce_upper_case=False,
        )

        end_point = self._get_price_history_end_point(
            symbol=symbol,
            time_window=time_window,
            output_size=output_size,
        )

        data = self._get_url(url=end_point)
        price_history = price_history_from_data(data=data)

        return price_history

    def _get_url(self, url: str) -> Dict[str, Any]:
        self._cool_off()
        data: Dict[str, Any] = json.loads(urlopen(url).read().decode('utf-8'))
        return data

    def _get_price_history_end_point(
        self,
        symbol: str,
        time_window: str,
        output_size: str,
    ) -> str:
        parameters = {
            'function': f'TIME_SERIES_{time_window}_ADJUSTED',
            'symbol': symbol,
            'outputsize': output_size,
            'apikey': self._get_api_key(),
        }

        return self._get_end_point(url_parameters=parameters)

    def _get_end_point(self, url_parameters: Dict[str, str]) -> str:
        return 'https://www.alphavantage.co/query?' + urlencode(url_parameters)

    def _get_cool_down_window(self) -> int:
        key = 'ALPHAVANTAGE_COOL_DOWN_WINDOW'
        cool_down_window = environ.get(key)
        if cool_down_window is None:
            raise ValueError(
                f'Failed to get the Alphavantage Cool Down Window from the '
                f'environment variable `{key}`.',
            )
        return int(cool_down_window)

    def _get_api_key(self) -> str:
        key = 'ALPHAVANTAGE_API_KEY'
        api_key = environ.get(key)
        if api_key is None:
            raise ValueError(
                f'Failed to get the Alphavantage API Key from the environment '
                f'variable `{key}`.',
            )

        return api_key

    def _check_str_parameter_in_range(
        self,
        parameter_name: str,
        parameter_value: str,
        allowed_values: Set[str],
        enforce_upper_case: bool,
    ) -> str:
        allowed_value_list: List[str] = list(allowed_values)
        if parameter_value.lower() not in allowed_values:
            allowed_values_str = ', '.join([
                f'`{value}`'
                for value in allowed_value_list[:-1]
            ])
            last_value = allowed_value_list[-1]
            allowed_values_str = f'{allowed_values_str} or `{last_value}`'
            raise ValueError(
                f'Invalid value for the parameter `{parameter_name}` - only '
                f'{allowed_values_str} is accepted.',
            )

        if enforce_upper_case:
            parameter_value = parameter_value.upper()
        else:
            parameter_value = parameter_value.lower()

        return parameter_value

    def _cool_off(self) -> None:
        if self._last_request_time is not None:
            time_elapsed = time() - self._last_request_time
            if time_elapsed < self._cool_down_window:
                sleep(self._cool_down_window - time_elapsed)
        self._last_request_time = time()


avc = AlphaVantageClient()



from copy import copy
from typing import Dict

from report.symbolhistory.symbolhistorycontextholding import SymbolHistoryContextHolding
from report.symbolhistory.symbolhistorycontextsnapshot import SymbolHistoryContextSnapshot
from robinhood.lib.marketplace.optioninstrument import OptionInstrument


class SymbolHistoryContext(object):

    def __init__(self, symbol: str) -> None:
        self._profit = 0.0
        self._stock_holding = SymbolHistoryContextHolding(
            symbol=symbol,
            quantity=0,
            avg_unit_price=0.0,
            option_instrument=None,
        )
        self._option_sell_holding: Dict[str, SymbolHistoryContextHolding] = {}
        self._option_buy_holding: Dict[str, SymbolHistoryContextHolding] = {}

    @property
    def stock_cost_basis(self) -> float:
        return self._stock_holding.quantity * self._stock_holding.avg_unit_price

    @property
    def option_sell_holding(self) -> Dict[str, SymbolHistoryContextHolding]:
        return self._option_sell_holding

    @property
    def option_buy_holding(self) -> Dict[str, SymbolHistoryContextHolding]:
        return self._option_buy_holding

    def take_snapshot(self) -> SymbolHistoryContextSnapshot:
        return SymbolHistoryContextSnapshot(
            profit=self._profit,
            stock_holding=self._stock_holding,
            option_sell_holding=copy(self._option_sell_holding),
            option_buy_holding=copy(self._option_buy_holding),
        )

    def limit_or_market_buy_stock(self, quantity: int, price: float) -> None:
        total_cost = self.stock_cost_basis + quantity * price
        total_quantity = self._stock_holding.quantity + quantity
        self._stock_holding = SymbolHistoryContextHolding(
            symbol=self._stock_holding.symbol,
            quantity=self._stock_holding.quantity + quantity,
            avg_unit_price=total_cost / total_quantity,
            option_instrument=None,
        )

    def limit_or_market_sell_stock(self, quantity: int, price: float) -> None:
        self._profit += (price - self._stock_holding.avg_unit_price) * quantity
        self._stock_holding = SymbolHistoryContextHolding(
            symbol=self._stock_holding.symbol,
            quantity=self._stock_holding.quantity - quantity,
            avg_unit_price=self._stock_holding.avg_unit_price,
            option_instrument=None,
        )

    def sell_to_open(
        self,
        symbol: str,
        quantity: int,
        unit_price: float,
        option_instrument: OptionInstrument,
    ) -> None:
        self._buy_or_sell_to_open(
            buy_or_sell='sell',
            symbol=symbol,
            quantity=quantity,
            unit_price=unit_price,
            option_instrument=option_instrument,
        )

    def buy_to_close(
        self,
        symbol: str,
        quantity: int,
        unit_price: float,
    ) -> None:
        self._buy_or_sell_to_close(
            buy_or_sell='buy',
            symbol=symbol,
            quantity=quantity,
            unit_price=unit_price,
        )

    def buy_to_open(
        self,
        symbol: str,
        quantity: int,
        unit_price: float,
        option_instrument: OptionInstrument,
    ) -> None:
        self._buy_or_sell_to_open(
            buy_or_sell='buy',
            symbol=symbol,
            quantity=quantity,
            unit_price=unit_price,
            option_instrument=option_instrument,
        )

    def sell_to_close(
        self,
        symbol: str,
        quantity: int,
        unit_price: float,
    ) -> None:
        self._buy_or_sell_to_close(
            buy_or_sell='sell',
            symbol=symbol,
            quantity=quantity,
            unit_price=unit_price,
        )

    def call_assignment(self, symbol: str, quantity: int, price: float) -> None:
        holding = self._option_sell_holding[symbol]
        current_avg_unit_price = holding.avg_unit_price
        new_quantity = holding.quantity - quantity
        if new_quantity == 0:
            del self._option_sell_holding[symbol]
        else:
            self._option_sell_holding[symbol] = SymbolHistoryContextHolding(
                symbol=symbol,
                quantity=new_quantity,
                avg_unit_price=holding.avg_unit_price,
                option_instrument=holding.option_instrument,
            )

        shares_sold = quantity * 100
        profit_per_share = price - self._stock_holding.avg_unit_price + current_avg_unit_price
        self._profit += profit_per_share * shares_sold
        self._stock_holding = SymbolHistoryContextHolding(
            symbol=self._stock_holding.symbol,
            quantity=self._stock_holding.quantity - shares_sold,
            avg_unit_price=self._stock_holding.avg_unit_price,
            option_instrument=None,
        )

    def put_asignment(self, symbol: str, quantity: int, price: float, strike_price: float) -> None:
        holding = self._option_sell_holding[symbol]
        new_quantity = holding.quantity - quantity
        if new_quantity == 0:
            del self._option_sell_holding[symbol]
        else:
            self._option_sell_holding[symbol] = SymbolHistoryContextHolding(
                symbol=symbol,
                quantity=new_quantity,
                avg_unit_price=holding.avg_unit_price,
                option_instrument=holding.option_instrument,
            )

        shares_bought = quantity * 100
        new_cost = (strike_price - holding.avg_unit_price) * shares_bought
        self._stock_holding = SymbolHistoryContextHolding(
            symbol=self._stock_holding.symbol,
            quantity=self._stock_holding.quantity + shares_bought,
            avg_unit_price=(self.stock_cost_basis + new_cost) / (self._stock_holding.quantity + shares_bought),
            option_instrument=None,
        )

    def call_exercise(
        self,
        symbol: str,
        quantity: int,
        price: float,
        strike_price: float,
    ) -> None:
        holding = self._option_buy_holding[symbol]
        shares_bought = quantity * 100
        effective_cost_per_share = strike_price + price
        new_cost = effective_cost_per_share * shares_bought
        total_cost = self.stock_cost_basis + new_cost

        self._stock_holding = SymbolHistoryContextHolding(
            symbol=self._stock_holding.symbol,
            quantity=self._stock_holding.quantity + shares_bought,
            avg_unit_price=total_cost / self._stock_holding.quantity,
            option_instrument=None,
        )

        new_quantity = holding.quantity - quantity
        if new_quantity == 0:
            del self._option_buy_holding[symbol]
        else:
            self._option_buy_holding[symbol] = SymbolHistoryContextHolding(
                symbol=symbol,
                quantity=new_quantity,
                avg_unit_price=holding.avg_unit_price,
                option_instrument=holding.option_instrument,
            )

    def option_expire(self, symbol: str, quantity: int) -> None:
        if symbol in self._option_buy_holding:
            holding_dict = self._option_buy_holding
            holding = self._option_buy_holding[symbol]
            profit = -holding.quantity * holding.avg_unit_price * 100
        elif symbol in self._option_sell_holding:
            holding_dict = self._option_sell_holding
            holding = self._option_sell_holding[symbol]
            profit = holding.quantity * holding.avg_unit_price * 100
        else:
            return

        assert holding.quantity - quantity == 0
        del holding_dict[symbol]
        self._profit += profit

    def _buy_or_sell_to_open(
        self,
        buy_or_sell: str,
        symbol: str,
        quantity: int,
        unit_price: float,
        option_instrument: OptionInstrument,
    ) -> None:
        if buy_or_sell == 'buy':
            option_holding_dict = self._option_buy_holding
        else:
            option_holding_dict = self._option_sell_holding

        if symbol in option_holding_dict:
            holding = option_holding_dict[symbol]
            new_quantity = holding.quantity + quantity
            new_price = (holding.quantity * holding.avg_unit_price + quantity * unit_price) / new_quantity
        else:
            new_quantity = quantity
            new_price = unit_price

        option_holding_dict[symbol] = SymbolHistoryContextHolding(
            symbol=symbol,
            quantity=new_quantity,
            avg_unit_price=new_price,
            option_instrument=option_instrument,
        )

    def _buy_or_sell_to_close(
        self,
        buy_or_sell: str,
        symbol: str,
        quantity: int,
        unit_price: float,
    ) -> None:
        if buy_or_sell == 'buy':
            option_holding_dict = self._option_sell_holding
            unit_profit = option_holding_dict[symbol].avg_unit_price - unit_price
        else:
            option_holding_dict = self._option_buy_holding
            unit_profit = unit_price - option_holding_dict[symbol].avg_unit_price

        new_quantity = option_holding_dict[symbol].quantity - quantity
        self._profit += quantity * unit_profit * 100
        if new_quantity == 0:
            del option_holding_dict[symbol]
        else:
            option_holding_dict[symbol] = SymbolHistoryContextHolding(
                symbol=symbol,
                quantity=new_quantity,
                avg_unit_price=option_holding_dict[symbol].avg_unit_price,
                option_instrument=option_holding_dict[symbol].option_instrument,
            )



from typing import Any
from typing import Dict
from typing import List

from robinhood.lib.account.optionorder.optionlegexecution import option_leg_execution_from_data
from robinhood.lib.account.optionorder.optionlegexecution import OptionLegExecution
from robinhood.lib.marketplace.optioninstrument import option_instrument_from_url
from robinhood.lib.marketplace.optioninstrument import OptionInstrument


class OptionLeg(object):

    def __init__(
        self,
        leg_id: str,
        option_instrument_url: str,
        side: str,
        position_effect: str,
        ratio_quantity: int,
        executions: List[OptionLegExecution],
    ) -> None:
        self._leg_id = leg_id
        self._option_instrument_url = option_instrument_url
        self._side = side
        self._position_effect = position_effect
        self._ratio_quantity = ratio_quantity
        self._executions = executions

    @property
    def option_instrument(self) -> OptionInstrument:
        return option_instrument_from_url(url=self._option_instrument_url)

    @property
    def is_call(self) -> bool:
        return self.option_instrument.is_call

    @property
    def is_put(self) -> bool:
        return self.option_instrument.is_put

    @property
    def quantity(self) -> float:
        total_quantity = 0.0
        for ex in self._executions:
            total_quantity += ex.quantity
        return total_quantity

    @property
    def ratio_quantity(self) -> int:
        return self._ratio_quantity

    @property
    def unit_price(self) -> float:
        total_price = 0.0
        for ex in self._executions:
            total_price += ex.quantity * ex.price
        return total_price / self.quantity

    @property
    def position_effect(self) -> str:
        return self._position_effect

    @property
    def is_open(self) -> bool:
        return self.position_effect == 'open'

    @property
    def is_close(self) -> bool:
        return self.position_effect == 'close'

    @property
    def side(self) -> str:
        return self._side

    @property
    def is_buy(self) -> bool:
        return self.side == 'buy'

    @property
    def is_sell(self) -> bool:
        return self.side == 'sell'

    @property
    def is_sell_to_open(self) -> bool:
        return self.is_sell and self.is_open

    @property
    def is_sell_to_close(self) -> bool:
        return self.is_sell and self.is_close

    @property
    def is_buy_to_open(self) -> bool:
        return self.is_buy and self.is_open

    @property
    def is_buy_to_close(self) -> bool:
        return self.is_buy and self.is_close

    @property
    def executions(self) -> List[OptionLegExecution]:
        return self._executions


def option_leg_from_data(data: Dict[str, Any]) -> OptionLeg:
    executions = [
        option_leg_execution_from_data(data=row)
        for row in data['executions']
    ]
    return OptionLeg(
        leg_id=data['id'],
        option_instrument_url=data['option'],
        side=data['side'],
        position_effect=data['position_effect'],
        ratio_quantity=int(data['ratio_quantity']),
        executions=executions,
    )



from data.watchlist import ALL
from robinhood.lib.marketplace.marketplace import marketplace


symbols = {'AAPL', 'MSFT'}


def test_get_equities_endpoint() -> None:
    endpoint = marketplace._get_equities_endpoint(symbols=symbols)
    assert endpoint in {
        'https://api.robinhood.com/quotes/?symbols=AAPL,MSFT',
        'https://api.robinhood.com/quotes/?symbols=MSFT,AAPL',
    }


def test_get_equities() -> None:
    equities = marketplace.get_equities(symbols=symbols)
    assert len(equities.keys()) == 2
    for symbol in equities.keys():
        assert symbol in symbols


def test_equity_instrument_ids_not_change() -> None:
    equities = marketplace.get_equities(symbols=set(ALL.keys()))
    for symbol, equity in equities.items():
        assert equity.instrument_id == ALL[symbol]


def test_get_option_chains() -> None:
    option_chain_ids = {
        'AAPL': 'cee01a93-626e-4ee6-9b04-60e2fd1392d1',
        'MSFT': '1ac71e01-0677-42c6-a490-1457980954f8',
    }
    equity_instrument_ids = [ALL[symbol] for symbol in symbols]
    option_chains = marketplace.get_option_chains(equity_instrument_ids=equity_instrument_ids)
    assert len(option_chains) == 2
    for symbol, option_chain in option_chains.items():
        assert symbol in symbols
        assert option_chain.chain_id == option_chain_ids[symbol]



from report.accountsummary.accountsummary import AccountSummary


def test_gen_reports(account_summary_report: AccountSummary) -> None:
    text = account_summary_report._gen_reports()
    for s in {
        'Stock Value',
        'Option Value',
        'Margin Used',
        'Account Value',
        '# of Symbols',
        'Profit on Stocks',
        'Profit on Options',
        'Profit on Call Writes',
        'Profit on Put Writes',
        'Symbol',
        '# of Shares',
        'Settled Profit',
        'Last Price',
        'Buying Power',
    }:
        assert s in text



from datetime import datetime
from datetime import timedelta
from typing import Callable
from typing import Optional

from pytz import timezone


LA_TZ_NAME = 'America/Los_Angeles'


def format_percentage(x: float) -> str:
    return '{:,.2%}'.format(x)


def format_price(x: float, empty_for_zero: bool = False) -> str:
    r = '${:,.2f}'.format(x)
    if empty_for_zero and r == '$0.00':
        return '-'

    return r


def format_optional_price(x: Optional[float]) -> str:
    if x is None:
        return '-'
    else:
        return format_price(x=x)


def format_quantity(x: float) -> str:
    return '{:,.0f}'.format(x)


def format_greek(x: float) -> str:
    return '{:,.2}'.format(x)


def format_optional_greek(x: Optional[float]) -> str:
    if x is None:
        return '-'
    else:
        return format_greek(x=x)


def attach_timezone(t: datetime) -> datetime:
    return t.astimezone(tz=timezone(LA_TZ_NAME))


def parse_date(s: str) -> datetime:
    return datetime.strptime(s, '%Y-%m-%d').astimezone(tz=timezone(LA_TZ_NAME))


def parse_datetime(s: str) -> datetime:
    return datetime.strptime(s, '%Y-%m-%dT%H:%M:%SZ').astimezone(tz=timezone(LA_TZ_NAME))


def parse_datetime_with_microsecond(s: str) -> datetime:
    return datetime.strptime(s, '%Y-%m-%dT%H:%M:%S.%fZ').astimezone(tz=timezone(LA_TZ_NAME))


def parse_datetime_with_optional_microsecond(s: str) -> datetime:
    try:
        return parse_datetime_with_microsecond(s)
    except ValueError:
        return parse_datetime(s)


def parse_optional_float(s: Optional[str]) -> Optional[float]:
    return s if s is None else float(s)


def parse_optional_int(s: Optional[str]) -> Optional[int]:
    return s if s is None else int(s)


def get_week(d: datetime) -> str:
    _d = d - timedelta(days=d.weekday())
    return _d.strftime('%Y-%m-%d')


def get_month(d: datetime) -> str:
    _d = d - timedelta(days=(d.day - 1))
    return _d.strftime('%Y-%m-%d')


def get_event_date_format_func(interval: str) -> Callable[[datetime], str]:
    if interval == 'Week':
        event_date_func = get_week
    elif interval == 'Month':
        event_date_func = get_month
    else:
        assert interval == 'All Time'

        def event_date_func(d: datetime) -> str:
            return '-'

    return event_date_func



from collections import defaultdict
from typing import Any
from typing import DefaultDict
from typing import Dict
from typing import List
from typing import Optional
from typing import Set

from robinhood.lib.marketplace.minticks import MinTicks
from robinhood.lib.marketplace.optioninstrument import option_instrument_from_data
from robinhood.lib.marketplace.optioninstrument import OptionInstrument
from robinhood.lib.robinhoodclient import rhc


class OptionChain(object):

    def __init__(
        self,
        chain_id: str,
        symbol: str,
        trade_value_multiplier: float,
        expiration_dates: List[str],
        min_ticks: MinTicks,
    ) -> None:
        self._chain_id = chain_id
        self._symbol = symbol
        self._trade_value_multiplier = trade_value_multiplier
        self._expiration_dates = expiration_dates
        self._min_ticks = min_ticks

        self._call_option_items: Optional[DefaultDict[str, List[OptionInstrument]]] = None
        self._put_option_items: Optional[DefaultDict[str, List[OptionInstrument]]] = None

    @property
    def chain_id(self) -> str:
        return self._chain_id

    @property
    def symbol(self) -> str:
        return self._symbol

    @property
    def call_option_items(self) -> DefaultDict[str, List[OptionInstrument]]:
        if self._call_option_items is None:
            self._call_option_items = self._get_option_instruments(
                expiration_dates=self._expiration_dates,
                option_type='call',
            )

        return self._call_option_items

    @property
    def put_option_items(self) -> DefaultDict[str, List[OptionInstrument]]:
        if self._put_option_items is None:
            self._put_option_items = self._get_option_instruments(
                expiration_dates=self._expiration_dates,
                option_type='put',
            )

        return self._put_option_items

    @property
    def expiration_dates(self) -> List[str]:
        return self._expiration_dates

    @property
    def min_ticks(self) -> MinTicks:
        return self._min_ticks

    def get_option_instruments(self, exp_date: str, option_type: str) -> List[OptionInstrument]:
        items = self._get_option_instruments(
            expiration_dates=[exp_date],
            option_type=option_type,
        )[exp_date]
        return sorted(items, key=lambda x: x.strike_price)

    def _get_option_instruments(
        self,
        expiration_dates: List[str],
        option_type: str,
    ) -> DefaultDict[str, List[OptionInstrument]]:
        assert self._is_valid_option_type(option_type=option_type)

        endpoint = self._get_option_chain_item_endpoint(
            expiration_dates=expiration_dates,
            option_type=option_type,
        )
        option_instruments = rhc.get_objects(
            url=endpoint,
            callback=option_instrument_from_data,
            result_key='results',
        )
        option_instrument_symbols: Set[str] = set()
        result: DefaultDict[str, List[OptionInstrument]] = defaultdict(list)
        for option_instrument in option_instruments:
            if option_instrument.symbol not in option_instrument_symbols:
                result[option_instrument.exp_date].append(option_instrument)
                option_instrument_symbols.add(option_instrument.symbol)

        return result

    def _get_option_chain_item_endpoint(
        self,
        expiration_dates: List[str],
        option_type: str,
    ) -> str:
        assert self._is_valid_option_type(option_type=option_type)

        comma_separated_exp_dates = ','.join(expiration_dates)
        state = 'active'
        tradability = 'tradable'
        return (
            f'https://api.robinhood.com/options/instruments/'
            f'?chain_id={self._chain_id}'
            f'&expiration_dates={comma_separated_exp_dates}'
            f'&type={option_type}'
            f'&state={state}'
            f'&tradability={tradability}'
        )

    def _is_valid_option_type(self, option_type: str) -> bool:
        return option_type in {'call', 'put'}


def option_chain_from_data(data: Dict[str, Any]) -> OptionChain:
    return OptionChain(
        chain_id=data['id'],
        symbol=data['symbol'],
        trade_value_multiplier=float(data['trade_value_multiplier']),
        expiration_dates=sorted(data['expiration_dates']),
        min_ticks=MinTicks(
            cutoff_price=float(data['min_ticks']['cutoff_price']),
            below_tick=float(data['min_ticks']['below_tick']),
            above_tick=float(data['min_ticks']['above_tick']),
        ),
    )



from collections import defaultdict
from typing import DefaultDict
from typing import List
from typing import Optional
from typing import Set
from typing import Tuple
from typing import TYPE_CHECKING

from pandas import DataFrame
from terminaltables import AsciiTable

from robinhood.lib.helpers import format_price
from robinhood.lib.helpers import get_event_date_format_func

if TYPE_CHECKING:
    from report.symbolhistory.symbolhistory import SymbolHistory


class AccountPerformance(object):

    def __init__(self, symbol_history: 'SymbolHistory') -> None:
        self._symbol_history = symbol_history

    def show(self, interval: str, symbol: Optional[str] = None) -> None:
        assert interval in {'Week', 'Month', 'All Time'}
        print(self._gen_reports(interval=interval, symbol=symbol))

    def _gen_reports(self, interval: str, symbol: Optional[str] = None) -> str:
        data = self._gen_data(interval=interval, symbol=symbol)
        t = AsciiTable(data)
        t.inner_column_border = True
        t.inner_footing_row_border = False
        t.inner_heading_row_border = True
        t.inner_row_border = False
        t.outer_border = False
        t.justify_columns = {
            0: 'center',
            1: 'right',
            2: 'right',
            3: 'right',
        }

        return str(t.table)

    def gen_df(self, interval: str) -> DataFrame:
        event_date_func = get_event_date_format_func(interval=interval)

        stock_profit: DefaultDict[str, DefaultDict[str, float]] = defaultdict(lambda: defaultdict(float))
        option_profit: DefaultDict[str, DefaultDict[str, float]] = defaultdict(lambda: defaultdict(float))

        for symbol, events in self._symbol_history.events_by_symbol.items():
            for i in range(1, len(events)):
                previous_event = events[i - 1]
                current_event = events[i]
                diff = current_event.context.profit - previous_event.context.profit
                if diff != 0:
                    event_date = event_date_func(current_event.event_ts)

                    if current_event.event_name in {
                        'BUY TO CLOSE',
                        'CUSTOM',
                        'CUSTOM CLOSE',
                        'FIG LEAF CLOSE',
                        'IRON CONDOR CLOSE',
                        'LONG CALL CLOSE',
                        'LONG CALL SPREAD CLOSE',
                        'LONG PUT CLOSE',
                        'LONG PUT SPREAD CLOSE',
                        'OPTION EXPIRATION',
                        'ROLLING SHORT CALL SPREAD',
                        'SHORT CALL SPREAD CLOSE',
                        'SHORT PUT SPREAD CLOSE',
                    }:
                        option_profit[symbol][event_date] += diff
                    elif current_event.event_name in {
                        'CALL ASSIGNMENT',
                        'LIMIT SELL',
                        'MARKET SELL',
                    }:
                        stock_profit[symbol][event_date] += diff
                    else:
                        raise Exception(f'Unrecognized event name: {current_event.event_name}')

        combined_data: DefaultDict[Tuple[str, str], Tuple[float, float]] = defaultdict(lambda: (0.0, 0.0))
        for symbol, stock_profit_by_symbol in stock_profit.items():
            for interval, profit in stock_profit_by_symbol.items():
                combined_data[(interval, symbol)] = (
                    combined_data[(interval, symbol)][0] + profit,
                    combined_data[(interval, symbol)][1],
                )

        for symbol, option_profit_by_symbol in option_profit.items():
            for interval, profit in option_profit_by_symbol.items():
                combined_data[(interval, symbol)] = (
                    combined_data[(interval, symbol)][0],
                    combined_data[(interval, symbol)][1] + profit,
                )

        data = []
        for (interval, symbol), (_stock_profit, _option_profit) in combined_data.items():
            data.append([
                interval,
                symbol,
                _stock_profit,
                _option_profit,
                _stock_profit + _option_profit,
            ])

        return DataFrame(
            data=data,
            columns=[
                'interval',
                'symbol',
                'stock_profit',
                'option_profit',
                'total_profit',
            ],
        )

    def _gen_data(self, interval: str, symbol: Optional[str] = None) -> List[List[str]]:
        event_date_func = get_event_date_format_func(interval=interval)

        data = [[
            interval,
            'Settled Profit from Stock',
            'Settled Profit from Option',
            'Total Settled Profit',
        ]]

        stock_profit: DefaultDict[str, float] = defaultdict(float)
        option_profit: DefaultDict[str, float] = defaultdict(float)

        for _symbol, events in self._symbol_history.events_by_symbol.items():
            if symbol is not None and symbol != _symbol:
                continue

            for i in range(1, len(events)):
                previous_event = events[i - 1]
                current_event = events[i]
                diff = current_event.context.profit - previous_event.context.profit
                if diff != 0:
                    event_date = event_date_func(current_event.event_ts)

                    if current_event.event_name in {
                        'BUY TO CLOSE',
                        'CUSTOM',
                        'CUSTOM CLOSE',
                        'FIG LEAF CLOSE',
                        'IRON CONDOR CLOSE',
                        'LONG CALL CLOSE',
                        'LONG CALL SPREAD CLOSE',
                        'LONG PUT CLOSE',
                        'LONG PUT SPREAD CLOSE',
                        'OPTION EXPIRATION',
                        'ROLLING SHORT CALL SPREAD',
                        'SHORT CALL SPREAD CLOSE',
                        'SHORT PUT SPREAD CLOSE',
                    }:
                        option_profit[event_date] += diff
                    elif current_event.event_name in {
                        'CALL ASSIGNMENT',
                        'LIMIT SELL',
                        'MARKET SELL',
                    }:
                        stock_profit[event_date] += diff
                    else:
                        raise Exception(f'Unrecognized event name: {current_event.event_name}')

        all_dates: Set[str] = set()
        for event_date in stock_profit.keys():
            all_dates.add(event_date)
        for event_date in option_profit.keys():
            all_dates.add(event_date)

        for event_date in sorted(list(all_dates)):
            total_profit = stock_profit[event_date] + option_profit[event_date]
            data.append([
                event_date,
                format_price(stock_profit[event_date]),
                format_price(option_profit[event_date]),
                format_price(total_profit),
            ])

        return data



from typing import Any
from typing import Optional

from robinhood.lib.helpers import parse_optional_float
from robinhood.lib.helpers import parse_optional_int


class OptionMarketData(object):

    def __init__(
        self,
        adjusted_mark_price: float,
        ask_price: float,
        ask_size: int,
        bid_price: float,
        bid_size: int,
        last_trade_price: Optional[float],
        last_trade_size: Optional[int],
        open_interest: int,
        implied_volatility: Optional[float],
        chance_of_profit_long: Optional[float],
        chance_of_profit_short: Optional[float],
        delta: Optional[float],
        gamma: Optional[float],
        rho: Optional[float],
        theta: Optional[float],
        vega: Optional[float],
        high_fill_rate_buy_price: Optional[float],
        high_fill_rate_sell_price: Optional[float],
        low_fill_rate_buy_price: Optional[float],
        low_fill_rate_sell_price: Optional[float],
    ) -> None:
        self._adjusted_mark_price = adjusted_mark_price
        self._ask_price = ask_price
        self._ask_size = ask_size
        self._bid_price = bid_price
        self._bid_size = bid_size
        self._last_trade_price = last_trade_price
        self._last_trade_size = last_trade_size
        self._open_interest = open_interest
        self._implied_volatility = implied_volatility
        self._chance_of_profit_long = chance_of_profit_long
        self._chance_of_profit_short = chance_of_profit_short
        self._delta = delta
        self._gamma = gamma
        self._rho = rho
        self._theta = theta
        self._vega = vega
        self._high_fill_rate_buy_price = high_fill_rate_buy_price
        self._high_fill_rate_sell_price = high_fill_rate_sell_price
        self._low_fill_rate_buy_price = low_fill_rate_buy_price
        self._low_fill_rate_sell_price = low_fill_rate_sell_price

    @property
    def adjusted_mark_price(self) -> float:
        return self._adjusted_mark_price

    @property
    def implied_volatility(self) -> Optional[float]:
        return self._implied_volatility

    @property
    def last_trade_price(self) -> Optional[float]:
        return self._last_trade_price

    @property
    def chance_of_profit_long(self) -> Optional[float]:
        return self._chance_of_profit_long

    @property
    def chance_of_profit_short(self) -> Optional[float]:
        return self._chance_of_profit_short

    @property
    def high_fill_rate_sell_price(self) -> Optional[float]:
        return self._high_fill_rate_sell_price

    @property
    def medium_fill_rate_sell_price(self) -> Optional[float]:
        if self._high_fill_rate_sell_price is None or self._low_fill_rate_sell_price is None:
            return None

        return (self._high_fill_rate_sell_price + self._low_fill_rate_sell_price) / 2

    @property
    def low_fill_rate_sell_price(self) -> Optional[float]:
        return self._low_fill_rate_sell_price

    @property
    def delta(self) -> Optional[float]:
        return self._delta

    @property
    def theta(self) -> Optional[float]:
        return self._theta

    @property
    def gamma(self) -> Optional[float]:
        return self._gamma

    @property
    def rho(self) -> Optional[float]:
        return self._rho

    @property
    def vega(self) -> Optional[float]:
        return self._vega

    @property
    def open_interest(self) -> int:
        return self._open_interest


def option_market_data_from_data(data: Any) -> OptionMarketData:
    return OptionMarketData(
        adjusted_mark_price=float(data['adjusted_mark_price']),
        ask_price=float(data['ask_price']),
        ask_size=int(data['ask_size']),
        bid_price=float(data['bid_price']),
        bid_size=int(data['bid_size']),
        last_trade_price=parse_optional_float(data['last_trade_price']),
        last_trade_size=parse_optional_int(data['last_trade_size']),
        open_interest=int(data['open_interest']),
        implied_volatility=parse_optional_float(data['implied_volatility']),
        chance_of_profit_long=parse_optional_float(data['chance_of_profit_long']),
        chance_of_profit_short=parse_optional_float(data['chance_of_profit_short']),
        delta=parse_optional_float(data['delta']),
        gamma=parse_optional_float(data['gamma']),
        rho=parse_optional_float(data['rho']),
        theta=parse_optional_float(data['theta']),
        vega=parse_optional_float(data['vega']),
        high_fill_rate_buy_price=parse_optional_float(data['high_fill_rate_buy_price']),
        high_fill_rate_sell_price=parse_optional_float(data['high_fill_rate_sell_price']),
        low_fill_rate_buy_price=parse_optional_float(data['low_fill_rate_buy_price']),
        low_fill_rate_sell_price=parse_optional_float(data['low_fill_rate_sell_price']),
    )



from concurrent.futures import as_completed
from concurrent.futures import Future
from concurrent.futures import ThreadPoolExecutor
from typing import Any
from typing import List
from typing import Optional
from typing import Set

from robinhood.lib.account.accountobj import account_obj_from_data
from robinhood.lib.account.accountobj import AccountObj
from robinhood.lib.account.optionevent.optionevent import OptionEvent
from robinhood.lib.account.optionorder.optionorder import OptionOrder
from robinhood.lib.account.optionposition.optionposition import OptionPosition
from robinhood.lib.account.stockorder.stockorder import StockOrder
from robinhood.lib.account.stockposition.stockposition import StockPosition
from robinhood.lib.robinhoodclient import rhc


class Account(object):

    ENDPOINT = 'https://api.robinhood.com/accounts/'

    def __init__(self) -> None:
        self._account_obj: Optional[AccountObj] = None
        self._stock_orders: Optional[List[StockOrder]] = None
        self._stock_positions: Optional[List[StockPosition]] = None
        self._option_orders: Optional[List[OptionOrder]] = None
        self._option_positions: Optional[List[OptionPosition]] = None
        self._option_events: Optional[List[OptionEvent]] = None

    @property
    def account_obj(self) -> AccountObj:
        if self._account_obj is None:
            self._account_obj = self._get_account_obj()

        return self._account_obj

    @property
    def margin_used(self) -> float:
        if self.account_obj.cash >= 0:
            margin_used = 0.0
        else:
            margin_used = -self.account_obj.cash

        return margin_used

    @property
    def buying_power(self) -> float:
        return self.account_obj.margin_balance.overnight_buying_power * 2

    @property
    def account_type(self) -> str:
        return self.account_obj.account_type

    @property
    def stock_orders(self) -> List[StockOrder]:
        if self._stock_orders is None:
            self._stock_orders = self.get_stock_orders()

        return self._stock_orders

    @property
    def stock_positions(self) -> List[StockPosition]:
        if self._stock_positions is None:
            self._stock_positions = self.get_stock_positions()

        return self._stock_positions

    @property
    def option_orders(self) -> List[OptionOrder]:
        if self._option_orders is None:
            self._option_orders = self.get_option_orders()

        return self._option_orders

    @property
    def option_positions(self) -> List[OptionPosition]:
        if self._option_positions is None:
            self._option_positions = self.get_option_positions()

        return self._option_positions

    @property
    def option_events(self) -> List[OptionEvent]:
        if self._option_events is None:
            self._option_events = self.get_option_events()

        return self._option_events

    @property
    def active_symbols(self) -> Set[str]:
        return self._get_active_symbols()

    def get_stock_orders(self) -> List[StockOrder]:
        return self.account_obj.get_stock_orders()

    def get_stock_positions(self) -> List[StockPosition]:
        return self.account_obj.get_stock_positions()

    def get_option_orders(self) -> List[OptionOrder]:
        return self.account_obj.get_option_orders()

    def get_option_positions(self) -> List[OptionPosition]:
        return self.account_obj.get_option_positions()

    def get_option_events(self) -> List[OptionEvent]:
        return self.account_obj.get_option_events()

    def update_in_parallel(
        self,
        update_stock_orders: bool = True,
        update_stock_positions: bool = True,
        update_option_orders: bool = True,
        update_option_positions: bool = True,
        update_option_events: bool = True,
    ) -> None:
        if self._account_obj is None:
            self._account_obj = self._get_account_obj()

        with ThreadPoolExecutor() as executor:
            futures: List[Future[List[Any]]] = []

            if update_stock_orders:
                futures.append(executor.submit(self.get_stock_orders))

            if update_stock_positions:
                futures.append(executor.submit(self.get_stock_positions))

            if update_option_orders:
                futures.append(executor.submit(self.get_option_orders))

            if update_option_positions:
                futures.append(executor.submit(self.get_option_positions))

            if update_option_events:
                futures.append(executor.submit(self.get_option_events))

        for future in as_completed(futures):
            results = future.result()
            if len(results) == 0:
                continue

            if isinstance(results[0], StockOrder):
                self._stock_orders = results
            elif isinstance(results[0], StockPosition):
                self._stock_positions = results
            elif isinstance(results[0], OptionOrder):
                self._option_orders = results
            elif isinstance(results[0], OptionPosition):
                self._option_positions = results
            elif isinstance(results[0], OptionEvent):
                self._option_events = results

    def _get_account_obj(self) -> AccountObj:
        account_obj: AccountObj = rhc.get_object(
            url=self.ENDPOINT,
            callback=account_obj_from_data,
            result_key='results',
        )
        return account_obj

    def _get_active_symbols(self) -> Set[str]:
        active_symbols: Set[str] = set()

        for stock_position in self.stock_positions:
            active_symbols.add(stock_position.symbol)

        for option_position in self.option_positions:
            if option_position.active_quantity > 0:
                active_symbols.add(option_position.chain_symbol)

        return active_symbols


account = Account()
